/*****************************************************************************/
/* jpcLinks.c                                                                */
/* Joe Christy	3/31/91                                                      */
/*****************************************************************************/
/*                                                                           */
/*    Copyright (C) 1991  Joe Christy                                        */
/*                                                                           */
/*    This program is free software; you can redistribute it and/or modify   */
/*    it under the terms of the GNU General Public License as published by   */
/*    the Free Software Foundation; either version 1, or (at your option)    */
/*    any later version.                                                     */
/*                                                                           */
/*    This program is distributed in the hope that it will be useful,        */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*    GNU General Public License for more details.                           */
/*                                                                           */
/*    You should have received a copy of the GNU General Public License      */
/*    along with this program; if not, write to the Free Software            */
/*    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */
/*                                                                           */
/*****************************************************************************/

#include "jpcLinks.h"

int             openComponents = 0,
                num_free_loops = 0,
                num_crossings = 0;
crossing_location *crossing_location_list = NULL;
component      *component_list = NULL;
edge           *current_edge = NULL;

static void     getLink(FILE * linkFile);
static void     create_old_projection(old_crossing **old_projection_ptr);
static int      permute_crossing_indices(old_crossing *old_projection_ptr);
static void     interchange_crossing_indices(old_crossing *old_projection_ptr,
							                 int *permuted, int i0, int i1);
static int      old_to_new_link(int num_free_loops, int num_crossings,
						        old_crossing *old_crossing_array,
								link_projection **link_ptr);
static int      find_dehydrated_link(char *link_name, char *input_string);
static void     clearLink();

/*
 * reconstituteLink() takes the place of rehydrate_link(). Returns 1 on
 * failure, 0 otherwise. This could probably be cleaner, but I couldn't
 * follow the logic of and relationship between pretty_pictures.c and
 * read_link.c
 * JPC	3/31/91 
 */
int             reconstituteLink(char *manifold_name,
							     triangulation **manifold_ptr, int mult)
{
	link_projection *link;
	char            path_name[PATH_NAME_LENGTH],
	                input_string[78];
	FILE           *fp;
	old_crossing   *old_projection;

	link = NULL;

	strcpy(path_name, LINK_DIRECTORY);
	strcat(path_name, manifold_name);
	fp = fopen(path_name, "r");
	if (!fp) {
		if (find_dehydrated_link(manifold_name, input_string) ||
			dehydrated_to_new_link(input_string, &link)) {
				return (0);
		} else {
			if (check_link(link) ||
				triangulate_link_complement(link, manifold_ptr)) {
					free_link_storage(link);
					free_manifold_storage(*manifold_ptr);
					*manifold_ptr = NULL;
					return (0);
			}
			free_link_storage(link);
			strcpy((*manifold_ptr)->name, manifold_name);
			if (standard_simplification(*manifold_ptr, 16) != 1)
				return (1);
		}
	} else {
		getLink(fp);
		printf("\n\nFinding structure of %s complement\n", manifold_name);
		create_old_projection(&old_projection);
		if (permute_crossing_indices(old_projection) == 0
			&& old_to_new_link(num_free_loops, num_crossings, old_projection,
							   &link) == 0
			&& check_link(link) == 0
			&& triangulate_link_complement(link, manifold_ptr) == 0) {
				strcpy((*manifold_ptr)->name, manifold_name);
				if (standard_simplification(*manifold_ptr, 16) != 1) {
					fclose(fp);
					clearLink();
					free32(old_projection);
					free_link_storage(link);
					return (1);
				}
		} else {
			fclose(fp);
			clearLink();
			free32(old_projection);
			free_link_storage(link);
			return (1);
		}
		fclose(fp);
		clearLink();
	}

	free32(old_projection);
	free_link_storage(link);
	free_manifold_storage(*manifold_ptr);
	*manifold_ptr = NULL;
	return (0);

}

/*
 * getLink() reads the link data structure from linkFile 
 *
 * contains code from doOpenPic() in snappea's link_entry.c 
 *
 * As it stands, this will only handle links w/o open components,
 * but that's all we allow the user to save. 
 */
static void     getLink(FILE * linkFile)
{
	int             i,
	                num_components,
	                num_edges,
	                anchor_index,
	                edge_index,
	                next_edge_index,
	                crossing_index,
	                h,
	                v,
	                overstrand_index,
	                understrand_index,
	                over_next_index,
	                under_next_index;
	component     **component_ptrs;
	edge          **edge_ptrs;
	crossing      **crossing_ptrs,
	               *over_cr,
	               *under_cr;
	crossing_location **cr_loc_ptrs,
	               *cr_loc;

	/* read header information */
	fscanf(linkFile, "%d%d%d%d", &num_components, &num_edges, &num_crossings,
		   &num_free_loops);

	/* note: the memory for each struct must be allocated separately so	 */
	/* that it can be freed separately as the user edits the link		 */

	/* allocate address tables */
	component_ptrs =
		(component **)malloc32((size_t) num_components * sizeof(component *));
	edge_ptrs = (edge **)malloc32((size_t) num_edges * sizeof(edge));
	crossing_ptrs =
		(crossing **)malloc32((size_t) 2 * num_crossings * sizeof(crossing *));
	cr_loc_ptrs =
		(crossing_location **)malloc32((size_t) num_crossings *
									   sizeof(crossing_location *));

	/* allocate data structures */
	for (i = num_components; --i >= 0;)
		component_ptrs[i] = (component *)malloc32((size_t) sizeof(component));
	for (i = num_edges; --i >= 0;)
		edge_ptrs[i] = (edge *)malloc32((size_t) sizeof(edge));
	for (i = 2 * num_crossings; --i >= 0;)
		crossing_ptrs[i] = (crossing *)malloc32((size_t) sizeof(crossing));
	for (i = num_crossings; --i >= 0;)
		cr_loc_ptrs[i] =
			(crossing_location *)malloc32((size_t) sizeof(crossing_location));

	/* set up components */
	component_list = component_ptrs[0];
	for (i = 0; i < num_components; i++) {
		fscanf(linkFile, "%d", &anchor_index);
		component_ptrs[i]->anchor = edge_ptrs[anchor_index];
		component_ptrs[i]->buoy = edge_ptrs[anchor_index];
		component_ptrs[i]->next =
			((i == num_components - 1) ? NULL : component_ptrs[i + 1]);
	}

	/* set up edges */
	for (i = num_edges; --i >= 0;) {
		fscanf(linkFile, "%d%d%d%d%d", &edge_index, &h, &v, &crossing_index,
			   &next_edge_index);
		edge_ptrs[edge_index]->start.h = h;
		edge_ptrs[edge_index]->start.v = v;
		edge_ptrs[edge_index]->crossing_list =
			(crossing_index == -1 ? NULL : crossing_ptrs[crossing_index]);
		if (edge_ptrs[edge_index]->crossing_list)
			edge_ptrs[edge_index]->crossing_list->prev = NULL;
		edge_ptrs[edge_index]->next = edge_ptrs[next_edge_index];
		edge_ptrs[next_edge_index]->prev = edge_ptrs[edge_index];
	}
	for (i = 0; i < num_edges; i++)	   /* This will fail for open components */
		edge_ptrs[i]->finish = edge_ptrs[i]->next->start;

	/* set up crossings and crossing_locations */
	crossing_location_list =
		(num_crossings ? cr_loc_ptrs[num_crossings - 1] : NULL);
	for (i = num_crossings; --i >= 0;) {
		/* note: it's essential to read the crossings in descending order	 */
		cr_loc = cr_loc_ptrs[i];
		over_cr = crossing_ptrs[2 * i];
		under_cr = crossing_ptrs[2 * i + 1];
		fscanf(linkFile, "%d%d%d%d%d%d%d%d%d%d%d",
			   &cr_loc->square.top,
			   &cr_loc->square.left,
			   &cr_loc->square.bottom,
			   &cr_loc->square.right,
			   &overstrand_index,
			   &understrand_index,
			   &over_cr->r_or_l,
			   &over_cr->dist,
			   &under_cr->dist,
			   &over_next_index,
			   &under_next_index
			);
		under_cr->other_edge =
			cr_loc->overstrand = edge_ptrs[overstrand_index];
		over_cr->other_edge =
			cr_loc->understrand = edge_ptrs[understrand_index];
		cr_loc->overcrossing = over_cr;
		cr_loc->undercrossing = under_cr;
		cr_loc->next = (i ? cr_loc_ptrs[i - 1] : NULL);
		/* when i == 0 we're at the end of the list */
		over_cr->over_or_under = over;
		under_cr->over_or_under = under;
		under_cr->r_or_l = over_cr->r_or_l;
		over_cr->next =
			((over_next_index == -1) ? NULL : crossing_ptrs[over_next_index]);
		under_cr->next =
			((under_next_index == -1) ? NULL : crossing_ptrs[under_next_index]);
		if (over_cr->next)
			over_cr->next->prev = over_cr;
		if (under_cr->next)
			under_cr->next->prev = under_cr;
		over_cr->index0 = i;
		under_cr->index0 = i;
		over_cr->location = cr_loc;
		under_cr->location = cr_loc;
	}
	openComponents = 0;

	/* clean up */
	free32(component_ptrs);
	free32(edge_ptrs);
	free32(crossing_ptrs);
	free32(cr_loc_ptrs);
	fclose(linkFile);

	return;
}


/*
 * create_old_projection() is needed by reconstituteLink() 
 *
 * basically lifted straight from link_entry.c 
 */
static void     create_old_projection(old_crossing **old_projection_ptr)
{
	component      *com;
	edge           *edge0,
	               *edge1;
	crossing       *cr;
	int             crossing_found;
	old_crossing   *old_cr;
	unsigned int    next_index;
	signed char     next_height;

	*old_projection_ptr =
		(old_crossing *)malloc32((size_t) num_crossings * sizeof(old_crossing));

	num_free_loops = 0;

	for (com = component_list; com; com = com->next) {
		crossing_found = 0;
		edge0 = com->anchor;
		do {
			for (cr = edge0->crossing_list; cr; cr = cr->next) {
				crossing_found = 1;
				old_cr = *old_projection_ptr + cr->index0;
				old_cr->l_or_r = (cr->r_or_l == right_handed);
				/* 'L' = 0, 'R' = 1 */
				if (cr->next) {
					next_index = cr->next->index0;
					next_height = (cr->next->over_or_under == over);
					/* 'D' = 0, 'U' = 1 */
				} else {
					for (edge1 = edge0->next;
						 !edge1->crossing_list;
						 edge1 = edge1->next);
					next_index = edge1->crossing_list->index0;
					next_height = (edge1->crossing_list->over_or_under == over);
				}
				if (cr->over_or_under == over) {
					old_cr->up_crossing = next_index;
					old_cr->up_height = next_height;
				} else {
					old_cr->down_crossing = next_index;
					old_cr->down_height = next_height;
				}
			}
			edge0 = edge0->next;
		} while (edge0 != com->anchor);
		if (crossing_found == 0)
			num_free_loops++;
	}

	return;
}
/*
 * permute_crossing_indices() is needed by reconstituteLink() 
 *
 * basically lifted straight from link_entry.c 
 */
/* mild goto spaghetti here */

/* The purpose of permute_crossing_indices() is to change some	 */
/* of the crossing indices in the link projection so that the	 */
/* cusp numbers assigned by triangulate_link_complement() match	 */
/* the order in which the link components were created.  Note	 */
/* that this routine depends on the details of the triangulation */
/* algorithm!  I.e. it's correct at the moment, but the cusp	 */
/* indices are not passed explicitly so if you were to			 */
/* substitute a different triangulation algorithm you'd most	 */
/* likely get incorrect cusp numbering.							 */
/* permute_crossing_indices() returns 0 upon success, 1 if		 */
/* there is a component with no overcorssings.					 */

static int      permute_crossing_indices(old_crossing *old_projection)
{
	int             com_count,
	               *permuted,
	                i;
	component      *com;
	edge           *edge0;
	crossing       *cr;

	/* count components */
	com_count = 0;
	for (com = component_list; com; com = com->next)
		com_count++;

	/* Maintain an array to record permuted		 */
	/* indices as a function of "true" indices.	 */
	permuted = (int *)malloc32((size_t) num_crossings * sizeof(int));
	for (i = num_crossings; --i >= 0;)
		permuted[i] = i;

	/* To avoid writing off the end of the old_projection array		 */
	/* we have to make sure there are at least as many components	 */
	/* as crossings.  One can prove this is the case whenever		 */
	/* num_free_loops is zero.										 */
	if (num_free_loops)
		goto error_message;

	/* For each component find an overcrossing on that component	 */
	/* and make it have the same index as the component itself.		 */
	for (com = component_list; com; com = com->next) {
		edge0 = com->anchor;
		do {
			for (cr = edge0->crossing_list; cr; cr = cr->next)
				if (cr->over_or_under == over) {
					interchange_crossing_indices(old_projection, permuted,
												  --com_count,
												 permuted[cr->index0]);
					goto end_of_outer_loop;
				}
			edge0 = edge0->next;
		} while (edge0 != com->anchor);

error_message:
		printf("\nThe components can't be correctly numbered because\n");
		printf("a component has no overcrossings.\n");
		printf("The link complement is not hyperbolic.\n\n");
		free32(permuted);
		return (1);

end_of_outer_loop:
		 /* move on to next component */ ;
	}

	free32(permuted);
	return (0);
}

/*
 * interchange_crossing_indices() is needed by reconstituteLink() interchange
 * crossing indices i0 and i1 in old_projection 
 *
 * basically lifted straight from link_entry.c 
 */

/* Note: this algorithm is inefficient in that it scans down	 */
/* the list of crossings for each and every interchange.  A		 */
/* faster algorithm would be to record the necessary swaps in	 */
/* a permutation table, and then do them all at once.  This		 */
/* would reduce the time complexity from quadratic time to		 */
/* linear time, but at the moment it hardly seems worth it.		 */

static void     interchange_crossing_indices(old_crossing *old_projection,
							                  int *permuted, int i0, int i1)
{
	int             i;
	old_crossing    temp;
	/* switch i0 and i1 */
	temp = old_projection[i0];
	old_projection[i0] = old_projection[i1];
	old_projection[i1] = temp;

	/* switch references to i0 and i1 in old_projection */
	for (i = num_crossings; --i >= 0;) {
		if (old_projection[i].up_crossing == i0)
			old_projection[i].up_crossing = i1;
		else if (old_projection[i].up_crossing == i1)
			old_projection[i].up_crossing = i0;

		if (old_projection[i].down_crossing == i0)
			old_projection[i].down_crossing = i1;
		else if (old_projection[i].down_crossing == i1)
			old_projection[i].down_crossing = i0;
	}

	/* update permutation array */
	for (i = num_crossings; --i >= 0;) {
		if (permuted[i] == i0)
			permuted[i] = i1;
		else if (permuted[i] == i1)
			permuted[i] = i0;
	}

	return;
}

/*
 * old_to_new_link() is needed by reconstituteLink() basically lifted straight
 * from read_link.c 
 */

/*
 * Returns 0 if everything is OK, 1 if the link projection contains free
 * loops.  Allocates storage for the link_projection.  Does NOT free storage
 * for old_crossing_array.  
 */

static int      old_to_new_link(int num_free_loops, int num_crossings,
						        old_crossing *old_crossing_array,
								link_projection **link_ptr)
{
	int             i;
	old_crossing   *this_old_crossing;
	crossing_info  *this_crossing,
	               *crossing;
	link_projection *link;
	if (num_free_loops) {
		printf("Link contains free loops.  No hyperbolic structure is possible.\n");
		*link_ptr = NULL;
		return (1);
	}
	link = (link_projection *)malloc32((size_t) sizeof(link_projection));
	link->name[0] = 0;
	link->number_of_crossings = num_crossings;
	link->crossing_array =
		crossing =
	(crossing_info *)malloc32((size_t) num_crossings * sizeof(crossing_info));
	/* set backward_strand indices to 2 to allow for error checking */
	for (i = 0; i < num_crossings; i++) {
		crossing[i].backward_0_strand = 2;
		crossing[i].backward_1_strand = 2;
	}

	for (i = 0; i < num_crossings; i++) {
		this_crossing = &crossing[i];
		this_old_crossing = &old_crossing_array[i];

		if (this_crossing->over_strand = this_old_crossing->l_or_r) {
			this_crossing->forward_0_crossing = this_old_crossing->down_crossing;
			if (this_old_crossing->down_height ==
			  old_crossing_array[this_old_crossing->down_crossing].l_or_r) {
				this_crossing->forward_0_strand = 1;
				crossing[this_crossing->forward_0_crossing].backward_1_crossing = i;
				crossing[this_crossing->forward_0_crossing].backward_1_strand = 0;
			} else {
				this_crossing->forward_0_strand = 0;
				crossing[this_crossing->forward_0_crossing].backward_0_crossing = i;
				crossing[this_crossing->forward_0_crossing].backward_0_strand = 0;
			}
			this_crossing->forward_1_crossing = this_old_crossing->up_crossing;
			if (this_old_crossing->up_height ==
				old_crossing_array[this_old_crossing->up_crossing].l_or_r) {
				this_crossing->forward_1_strand = 1;
				crossing[this_crossing->forward_1_crossing].backward_1_crossing = i;
				crossing[this_crossing->forward_1_crossing].backward_1_strand = 1;
			} else {
				this_crossing->forward_1_strand = 0;
				crossing[this_crossing->forward_1_crossing].backward_0_crossing = i;
				crossing[this_crossing->forward_1_crossing].backward_0_strand = 1;
			}
		} else {
			this_crossing->forward_0_crossing = this_old_crossing->up_crossing;
			if (this_old_crossing->up_height ==
				old_crossing_array[this_old_crossing->up_crossing].l_or_r) {
				this_crossing->forward_0_strand = 1;
				crossing[this_crossing->forward_0_crossing].backward_1_crossing = i;
				crossing[this_crossing->forward_0_crossing].backward_1_strand = 0;
			} else {
				this_crossing->forward_0_strand = 0;
				crossing[this_crossing->forward_0_crossing].backward_0_crossing = i;
				crossing[this_crossing->forward_0_crossing].backward_0_strand = 0;
			}
			this_crossing->forward_1_crossing = this_old_crossing->down_crossing;
			if (this_old_crossing->down_height ==
			  old_crossing_array[this_old_crossing->down_crossing].l_or_r) {
				this_crossing->forward_1_strand = 1;
				crossing[this_crossing->forward_1_crossing].backward_1_crossing = i;
				crossing[this_crossing->forward_1_crossing].backward_1_strand = 1;
			} else {
				this_crossing->forward_1_strand = 0;
				crossing[this_crossing->forward_1_crossing].backward_0_crossing = i;
				crossing[this_crossing->forward_1_crossing].backward_0_strand = 1;
			}
		}
	}

	/* error checking */
	/*
	 * check that no two strands go to the same place (by checking that every
	 * backwards strand gets hit) 
	 */
	for (i = 0; i < num_crossings; i++)
		if (crossing[i].backward_0_strand == 2 || crossing[i].backward_1_strand == 2){
			printf("Inconsistent crossing data in old_to_new_link()\n");
			exit(0);
		}
	*link_ptr = link;
	return (0);
}

/* note: the following static variables also appear in */
/* main_rehydrate_links.c and read_link.c */
static int      max_components = 4;
static int      max_crossings[5] = {0, 10, 9, 9, 9};
static int      num_knots[5][11] = {
									{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
									{0, 0, 0, 1, 1, 2, 3, 7, 21, 49, 166},
									{1, 0, 1, 0, 1, 1, 3, 8, 16, 61, 0},
									{0, 0, 0, 0, 0, 0, 3, 1, 10, 21, 0},
									{0, 0, 0, 0, 0, 0, 0, 0, 3, 1, 0}};

static int      cumulative_count[5][11] = {
										   {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
									   {0, 0, 0, 0, 1, 2, 4, 7, 14, 35, 84},
						  {250, 0, 251, 0, 252, 253, 254, 257, 265, 281, 0},
								  {0, 0, 0, 0, 0, 0, 342, 345, 346, 356, 0},
									 {0, 0, 0, 0, 0, 0, 0, 0, 377, 380, 0}};

static int      find_dehydrated_link(char *link_name, char *input_string)
{
	int             i,
	                num_components,
	                num_crossings,
	                index;
	FILE           *fp;

	/* do we have an L followed by exactly six digits? */
	if (link_name[0] != 'l' && link_name[0] != 'L')
		return (-1);
	for (i = 1; i <= 6; i++)
		if (link_name[i] < '0' || link_name[i] > '9')
			return (-1);
	if (link_name[7])
		return (-1);

	/* do we have a valid link name? */
	num_components = (link_name[1] - '0');
	num_crossings = 10 * (link_name[2] - '0') + (link_name[3] - '0');
	index = 100 * (link_name[4] - '0') + 10 * (link_name[5] - '0') + (link_name[6] - '0');
	if (num_components > max_components)
		return (-1);
	if (num_crossings > max_crossings[num_components])
		return (-1);
	if (index == 0 || index > num_knots[num_components][num_crossings])
		return (-1);

	/* fish the appropriate string out of links.dehydrated */
	fp = fopen("links.dehydrated", "r");
	if (fp == NULL) {
		char            scratch[51];
		strcpy(scratch, DEHYDRATED_DIRECTORY);
		strcat(scratch, "links.dehydrated");
		fp = fopen(scratch, "r");
	}
	if (fp == NULL) {
		printf("couldn't open links.dehydrated to rehydrate %s\n", link_name);
		return (-1);
	}
	for (i = cumulative_count[num_components][num_crossings] + index; --i >= 0;)
		fscanf(fp, "%s", input_string);
	fclose(fp);

	return (0);
}

/*
 * clearLink() clears the internal data structures associated with a link
 * contains code from redraw_knot() in snappea's link_entry.c 
 */
static void     clearLink()
{
	crossing_location *next_crossing_location;
	component      *next_component;
	edge           *edge0,
	               *next_edge;
	int             componentOpen;
	while (crossing_location_list) {
		next_crossing_location = crossing_location_list->next;
		free32(crossing_location_list->overcrossing);
		free32(crossing_location_list->undercrossing);
		free32(crossing_location_list);
		crossing_location_list = next_crossing_location;
	}

	while (component_list) {
		componentOpen = (component_list->anchor != component_list->buoy);
		next_component = component_list->next;
		edge0 = component_list->anchor;
		if (!edge0->next) {			   /* we're on an open one edge component */
			free32(edge0);
		} else {
			do {
				next_edge = edge0->next;
				free32(edge0);
				edge0 = next_edge;
			} while (edge0 != component_list->buoy);
		}
		if (componentOpen)
			free32(edge0);
		free32(component_list);
		component_list = next_component;
	}

	component_list = NULL;
	openComponents = 0;
	num_free_loops = 0;
	num_crossings = 0;
	current_edge = NULL;

	return;
}
